MODULE limdyn
   !!======================================================================
   !!                     ***  MODULE  limdyn  ***
   !!   Sea-Ice dynamics :  
   !!======================================================================
#if defined key_lim3
   !!----------------------------------------------------------------------
   !!   'key_lim3' :                                 LIM3 sea-ice model
   !!----------------------------------------------------------------------
   !!    lim_dyn      : computes ice velocities
   !!    lim_dyn_init : initialization and namelist read
   !!----------------------------------------------------------------------
   !! * Modules used
   USE phycst
   USE in_out_manager  ! I/O manager
   USE dom_ice
   USE dom_oce         ! ocean space and time domain
   USE taumod
   USE ice
   USE par_ice
   USE ice_oce
   USE iceini
   USE limistate
   USE limrhg          ! ice rheology
   USE lbclnk
   USE lib_mpp
   USE prtctl          ! Print control

   IMPLICIT NONE
   PRIVATE

   !! * Accessibility
   PUBLIC lim_dyn  ! routine called by ice_step

   !! * Module variables
   REAL(wp)  ::  rone    = 1.e0   ! constant value

   !!----------------------------------------------------------------------
   !!   LIM 3.0,  UCL-ASTR-LOCEAN-IPSL (2008)
   !! $Header: /home/opalod/NEMOCVSROOT/NEMO/LIM_SRC/limdyn.F90,v 1.5 2005/03/27 18:34:41 opalod Exp $
   !! This software is governed by the CeCILL licence see modipsl/doc/NEMO_CeCILL.txt
   !!----------------------------------------------------------------------

CONTAINS

   SUBROUTINE lim_dyn
      !!-------------------------------------------------------------------
      !!               ***  ROUTINE lim_dyn  ***
      !!               
      !! ** Purpose :   compute ice velocity and ocean-ice stress
      !!                
      !! ** Method  : 
      !!
      !! ** Action  : - Initialisation
      !!              - Call of the dynamic routine for each hemisphere
      !!              - computation of the stress at the ocean surface         
      !!              - treatment of the case if no ice dynamic
      !! History :
      !!   1.0  !  01-04   (LIM)  Original code
      !!   2.0  !  02-08   (C. Ethe, G. Madec)  F90, mpp
      !!   3.0  !  2007-03 (M.A. Morales Maqueda, S. Bouillon, M. Vancoppenolle) 
      !!                   LIM3, EVP, C-grid
      !!------------------------------------------------------------------------------------
      !! * Local variables
      INTEGER ::   ji, jj, jl, ja    ! dummy loop indices
      INTEGER :: i_j1, i_jpj         ! Starting/ending j-indices for rheology

      REAL(wp) ::   &
         ztairx, ztairy,          &  ! tempory scalars
         zsang , zmod,            &
         ztglx , ztgly ,          &
         zt11, zt12, zt21, zt22 , &
         zustm,                   &
         zsfrldmx2, zsfrldmy2,    &
         zu_ice, zv_ice, ztair2

      REAL(wp),DIMENSION(jpj) ::   &
           zind,                     &  ! i-averaged indicator of sea-ice
           zmsk                         ! i-averaged of tmask
      !!---------------------------------------------------------------------

      WRITE(numout,*) ' lim_dyn : Ice dynamics '
      WRITE(numout,*) ' ~~~~~~~ '

      IF( numit == nstart  )   CALL lim_dyn_init   ! Initialization (first time-step only)
      
      IF ( ln_limdyn ) THEN

         ! ocean velocity
         u_oce(:,:)  = u_io(:,:) * tmu(:,:)
         v_oce(:,:)  = v_io(:,:) * tmv(:,:)
         
         old_u_ice(:,:) = u_ice(:,:) * tmu(:,:)
         old_v_ice(:,:) = v_ice(:,:) * tmv(:,:)

         ! Rheology (ice dynamics)
         ! ========

         !  Define the j-limits where ice rheology is computed
         ! ---------------------------------------------------

         IF( lk_mpp ) THEN                    ! mpp: compute over the whole domain
            i_j1 = 1
            i_jpj = jpj
            IF(ln_ctl) CALL prt_ctl_info( 'lim_dyn  :    i_j1 = ', ivar1=i_j1, clinfo2=' ij_jpj = ', ivar2=i_jpj )
            CALL lim_rhg( i_j1, i_jpj )
         ELSE                                 ! optimization of the computational area

            DO jj = 1, jpj
               zind(jj) = SUM( 1.0 - at_i (:,jj  ) )   ! = FLOAT(jpj) if ocean everywhere on a j-line
               zmsk(jj) = SUM( tmask(:,jj,1) )   ! = 0          if land  everywhere on a j-line
            END DO

            IF( l_jeq ) THEN                     ! local domain include both hemisphere
               !                                 ! Rheology is computed in each hemisphere
               !                                 ! only over the ice cover latitude strip
               ! Northern hemisphere
               i_j1  = njeq
               i_jpj = jpj
               DO WHILE ( i_j1 <= jpj .AND. zind(i_j1) == FLOAT(jpi) .AND. zmsk(i_j1) /=0 )
                  i_j1 = i_j1 + 1
               END DO
               i_j1 = MAX( 1, i_j1-1 )
               IF(ln_ctl) CALL prt_ctl_info( 'lim_dyn  : NH  i_j1 = ', ivar1=i_j1, clinfo2=' ij_jpj = ', ivar2=i_jpj )
               CALL lim_rhg( i_j1, i_jpj )

               ! Southern hemisphere
               i_j1  =  1
               i_jpj = njeq
               DO WHILE ( i_jpj >= 1 .AND. zind(i_jpj) == FLOAT(jpi) .AND. zmsk(i_jpj) /=0 )
                  i_jpj = i_jpj - 1
               END DO
               i_jpj = MIN( jpj, i_jpj+2 )
               IF(ln_ctl) CALL prt_ctl_info( 'lim_dyn  : SH  i_j1 = ', ivar1=i_j1, clinfo2=' ij_jpj = ', ivar2=i_jpj )

               CALL lim_rhg( i_j1, i_jpj )

            ELSE                                 ! local domain extends over one hemisphere only
               !                                 ! Rheology is computed only over the ice cover
               !                                 ! latitude strip
               i_j1  = 1
               DO WHILE ( i_j1 <= jpj .AND. zind(i_j1) == FLOAT(jpi) .AND. zmsk(i_j1) /=0 )
                  i_j1 = i_j1 + 1
               END DO
               i_j1 = MAX( 1, i_j1-1 )

               i_jpj  = jpj
               DO WHILE ( i_jpj >= 1  .AND. zind(i_jpj) == FLOAT(jpi) .AND. zmsk(i_jpj) /=0 )
                  i_jpj = i_jpj - 1
               END DO
               i_jpj = MIN( jpj, i_jpj+2)

               IF(ln_ctl) CALL prt_ctl_info( 'lim_dyn  : one hemisphere:  i_j1 = ', ivar1=i_j1, clinfo2=' ij_jpj = ', ivar2=i_jpj )

               CALL lim_rhg( i_j1, i_jpj )

            ENDIF

         ENDIF

         u_ice(:,1) = 0.0       !ibug  est-ce vraiment necessaire?
         v_ice(:,1) = 0.0

         IF(ln_ctl) THEN 
            CALL prt_ctl(tab2d_1=u_oce , clinfo1=' lim_dyn  : u_oce :', tab2d_2=v_oce , clinfo2=' v_oce :')
            CALL prt_ctl(tab2d_1=u_ice , clinfo1=' lim_dyn  : u_ice :', tab2d_2=v_ice , clinfo2=' v_ice :')
         ENDIF

         ! Ice-Ocean stress
         ! ================
         DO jj = 2, jpjm1
            zsang  = SIGN(1.e0, gphif(1,jj-1) ) * sangvg

            DO ji = 2, jpim1
               ! computation of wind stress over ocean in X and Y direction
#if defined key_coupled && defined key_lim_cp1
!              ztairx =  ( 1.0 - at_i(ji-1,jj)   ) * gtaux(ji-1,jj)   + &
!                        ( 1.0 - at_i(ji,jj)     ) * gtaux(ji,jj  )   + &
!                        ( 1.0 - at_i(ji-1,jj-1) ) * gtaux(ji-1,jj-1) + & 
!                        ( 1.0 - at_i(ji,jj-1)   ) * gtaux(ji,jj-1)

!              ztairy =  ( 1.0 - at_i(ji-1,jj)   ) * gtauy(ji-1,jj  ) + &
!                        ( 1.0 - at_i(ji,jj  )   ) * gtauy(ji,jj    ) + &
!                        ( 1.0 - at_i(ji-1,jj-1) ) * gtauy(ji-1,jj-1) + & 
!                        ( 1.0 - at_i(ji,jj-1)   ) * gtauy(ji,jj-1)
#else
               ztairx =  ( 2.0 - at_i(ji,jj) - at_i(ji+1,jj) ) * gtaux(ji,jj) / cai * cao
               ztairy =  ( 2.0 - at_i(ji,jj) - at_i(ji,jj+1) ) * gtauy(ji,jj) / cai * cao

               zsfrldmx2 = at_i(ji,jj) + at_i(ji+1,jj)
               zsfrldmy2 = at_i(ji,jj) + at_i(ji,jj+1)

#endif
               zu_ice   = u_ice(ji,jj) - u_oce(ji,jj)
               zv_ice   = v_ice(ji,jj) - v_oce(ji,jj)
               zmod     = SQRT( zu_ice * zu_ice + zv_ice * zv_ice ) 

               ! quadratic drag formulation
               ztglx   = zsfrldmx2 * rhoco * zmod * ( cangvg * zu_ice - zsang * zv_ice ) 
               ztgly   = zsfrldmy2 * rhoco * zmod * ( cangvg * zv_ice + zsang * zu_ice ) 
!
!              ! IMPORTANT
!              ! these lignes are bound to prevent numerical oscillations
!              ! in the ice-ocean stress
!              ! They are physically ill-based. There is a cleaner solution
!              ! to try (remember discussion in Paris Gurvan)
!
               ztglx   = ztglx * exp( - zmod / 0.5 )
               ztgly   = ztglx * exp( - zmod / 0.5 )

               tio_u(ji,jj) = - ( ztairx + 1.0 * ztglx ) / ( 2. * rau0 )
               tio_v(ji,jj) = - ( ztairy + 1.0 * ztgly ) / ( 2. * rau0 )
            END DO
         END DO
         
         ! computation of friction velocity
         DO jj = 2, jpjm1
            DO ji = 2, jpim1

               zu_ice   = u_ice(ji,jj) - u_io(ji,jj)
               zt11  = rhoco * zu_ice * zu_ice

               zu_ice   = u_ice(ji-1,jj) - u_io(ji-1,jj)
               zt12  = rhoco * zu_ice * zu_ice

               zv_ice   = v_ice(ji,jj) - v_io(ji,jj)
               zt21  = rhoco * zv_ice * zv_ice

               zv_ice   = v_ice(ji,jj-1) - v_io(ji,jj-1)
               zt22  = rhoco * zv_ice * zv_ice
               ztair2 = ( ( gtaux(ji,jj) + gtaux(ji-1,jj) ) / 2. )**2 + &
                        ( ( gtauy(ji,jj) + gtauy(ji,jj-1) ) / 2. )**2

               ! should not be weighted
               zustm =  ( at_i(ji,jj)       ) * 0.5 * ( zt11 + zt12 + zt21 + zt22 )        &
                  &  +  ( 1.0 - at_i(ji,jj) ) * SQRT( ztair2 )

               ust2s(ji,jj) = ( zustm / rau0 ) * ( rone + sdvt(ji,jj) ) * tms(ji,jj)

            END DO
         END DO

       ELSE      ! If no ice dynamics  
                    
               ! virer ca (key_lim_cp1)
          DO jj = 2, jpjm1
             DO ji = 2, jpim1
#if defined key_coupled && defined key_lim_cp1
                tio_u(ji,jj) = - (  gtaux(ji  ,jj  ) + gtaux(ji-1,jj  )       &
                   &              + gtaux(ji-1,jj-1) + gtaux(ji  ,jj-1) ) / ( 4 * rau0 )

                tio_v(ji,jj) = - (  gtauy(ji  ,jj )  + gtauy(ji-1,jj  )       &
                   &              + gtauy(ji-1,jj-1) + gtauy(ji  ,jj-1) ) / ( 4 * rau0 )
#else
                tio_u(ji,jj) = - gtaux(ji,jj) / cai * cao / rau0
                tio_v(ji,jj) = - gtauy(ji,jj) / cai * cao / rau0 
#endif
                ztair2 = ( ( gtaux(ji,jj) + gtaux(ji-1,jj) ) / 2. )**2 + &
                         ( ( gtauy(ji,jj) + gtauy(ji,jj-1) ) / 2. )**2
                zustm        = SQRT( ztair2  )

                ust2s(ji,jj) = ( zustm / rau0 ) * ( rone + sdvt(ji,jj) ) * tms(ji,jj)
            END DO
         END DO

      ENDIF

      CALL lbc_lnk( ust2s, 'T',  1. )   ! T-point
      CALL lbc_lnk( tio_u, 'U', -1. )   ! I-point (i.e. ice U-V point)
      CALL lbc_lnk( tio_v, 'V', -1. )   ! I-point (i.e. ice U-V point)

      IF(ln_ctl) THEN   ! Control print
         CALL prt_ctl(tab2d_1=tio_u     , clinfo1=' lim_dyn  : tio_u     :', tab2d_2=tio_v , clinfo2=' tio_v :')
         CALL prt_ctl(tab2d_1=ust2s     , clinfo1=' lim_dyn  : ust2s     :')
         CALL prt_ctl(tab2d_1=divu_i    , clinfo1=' lim_dyn  : divu_i    :')
         CALL prt_ctl(tab2d_1=delta_i   , clinfo1=' lim_dyn  : delta_i   :')
         CALL prt_ctl(tab2d_1=strength  , clinfo1=' lim_dyn  : strength  :')
         CALL prt_ctl(tab2d_1=area      , clinfo1=' lim_dyn  : cell area :')
         CALL prt_ctl(tab2d_1=at_i      , clinfo1=' lim_dyn  : at_i      :')
         CALL prt_ctl(tab2d_1=vt_i      , clinfo1=' lim_dyn  : vt_i      :')
         CALL prt_ctl(tab2d_1=vt_s      , clinfo1=' lim_dyn  : vt_s      :')
         CALL prt_ctl(tab2d_1=stress1_i , clinfo1=' lim_dyn  : stress1_i :')
         CALL prt_ctl(tab2d_1=stress2_i , clinfo1=' lim_dyn  : stress2_i :')
         CALL prt_ctl(tab2d_1=stress12_i, clinfo1=' lim_dyn  : stress12_i:')
         DO jl = 1, jpl
            CALL prt_ctl_info(' - Category : ', ivar1=jl)
            CALL prt_ctl_info('   ~~~~~~~~~~')
            CALL prt_ctl(tab2d_1=a_i   (:,:,jl)   , clinfo1= ' lim_dyn  : a_i      : ')
            CALL prt_ctl(tab2d_1=ht_i  (:,:,jl)   , clinfo1= ' lim_dyn  : ht_i     : ')
            CALL prt_ctl(tab2d_1=ht_s  (:,:,jl)   , clinfo1= ' lim_dyn  : ht_s     : ')
            CALL prt_ctl(tab2d_1=v_i   (:,:,jl)   , clinfo1= ' lim_dyn  : v_i      : ')
            CALL prt_ctl(tab2d_1=v_s   (:,:,jl)   , clinfo1= ' lim_dyn  : v_s      : ')
            CALL prt_ctl(tab2d_1=e_s   (:,:,1,jl) , clinfo1= ' lim_dyn  : e_s      : ')
            CALL prt_ctl(tab2d_1=t_su  (:,:,jl)   , clinfo1= ' lim_dyn  : t_su     : ')
            CALL prt_ctl(tab2d_1=t_s   (:,:,1,jl) , clinfo1= ' lim_dyn  : t_snow   : ')
            CALL prt_ctl(tab2d_1=sm_i  (:,:,jl)   , clinfo1= ' lim_dyn  : sm_i     : ')
            CALL prt_ctl(tab2d_1=smv_i (:,:,jl)   , clinfo1= ' lim_dyn  : smv_i    : ')
            DO ja = 1, nlay_i
               CALL prt_ctl_info(' - Layer : ', ivar1=ja)
               CALL prt_ctl_info('   ~~~~~~~')
               CALL prt_ctl(tab2d_1=t_i(:,:,ja,jl) , clinfo1= ' lim_dyn  : t_i      : ')
               CALL prt_ctl(tab2d_1=e_i(:,:,ja,jl) , clinfo1= ' lim_dyn  : e_i      : ')
            END DO
         END DO
      ENDIF

   END SUBROUTINE lim_dyn

    SUBROUTINE lim_dyn_init
      !!-------------------------------------------------------------------
      !!                  ***  ROUTINE lim_dyn_init  ***
      !!
      !! ** Purpose : Physical constants and parameters linked to the ice
      !!      dynamics
      !!
      !! ** Method  :  Read the namicedyn namelist and check the ice-dynamic
      !!       parameter values called at the first timestep (nit000)
      !!
      !! ** input   :   Namelist namicedyn
      !!
      !! history :
      !!  8.5  ! 03-08 (C. Ethe) original code
      !!  9.0  ! 07-03 (MA Morales Maqueda, S. Bouillon, M. Vancoppenolle)
      !!               EVP-Cgrid-LIM3
      !!-------------------------------------------------------------------
      NAMELIST/namicedyn/ epsd, alpha,     &
         &                dm, nbiter, nbitdr, om, resl, cw, angvg, pstar,   &
         &                c_rhg, etamn, creepl, ecc, ahi0, &
         &                nevp, telast, alphaevp
      !!-------------------------------------------------------------------

      ! Define the initial parameters
      ! -------------------------

      ! Read Namelist namicedyn
      REWIND ( numnam_ice )
      READ   ( numnam_ice  , namicedyn )
      IF(lwp) THEN
         WRITE(numout,*)
         WRITE(numout,*) 'lim_dyn_init : ice parameters for ice dynamics '
         WRITE(numout,*) '~~~~~~~~~~~~'
         WRITE(numout,*) '   tolerance parameter                              epsd   = ', epsd
         WRITE(numout,*) '   coefficient for semi-implicit coriolis           alpha  = ', alpha
         WRITE(numout,*) '   diffusion constant for dynamics                  dm     = ', dm
         WRITE(numout,*) '   number of sub-time steps for relaxation          nbiter = ', nbiter
         WRITE(numout,*) '   maximum number of iterations for relaxation      nbitdr = ', nbitdr
         WRITE(numout,*) '   relaxation constant                              om     = ', om
         WRITE(numout,*) '   maximum value for the residual of relaxation     resl   = ', resl
         WRITE(numout,*) '   drag coefficient for oceanic stress              cw     = ', cw
         WRITE(numout,*) '   turning angle for oceanic stress                 angvg  = ', angvg
         WRITE(numout,*) '   first bulk-rheology parameter                    pstar  = ', pstar
         WRITE(numout,*) '   second bulk-rhelogy parameter                    c_rhg  = ', c_rhg
         WRITE(numout,*) '   minimun value for viscosity                      etamn  = ', etamn
         WRITE(numout,*) '   creep limit                                      creepl = ', creepl
         WRITE(numout,*) '   eccentricity of the elliptical yield curve       ecc    = ', ecc
         WRITE(numout,*) '   horizontal diffusivity coeff. for sea-ice        ahi0   = ', ahi0
         WRITE(numout,*) '   number of iterations for subcycling              nevp   = ', nevp
         WRITE(numout,*) '   timescale for elastic waves                      telast = ', telast
         WRITE(numout,*) '   coefficient for the solution of int. stresses  alphaevp = ', alphaevp

      ENDIF

      usecc2 = 1.0 / ( ecc * ecc )
      rhoco  = rau0 * cw
      angvg  = angvg * rad
      sangvg = SIN( angvg )
      cangvg = COS( angvg )
      pstarh = pstar / 2.0

      !  Diffusion coefficients.
      ahiu(:,:) = ahi0 * umask(:,:,1)
      ahiv(:,:) = ahi0 * vmask(:,:,1)

   END SUBROUTINE lim_dyn_init

#else
   !!----------------------------------------------------------------------
   !!   Default option          Empty module           NO LIM sea-ice model
   !!----------------------------------------------------------------------
CONTAINS
   SUBROUTINE lim_dyn         ! Empty routine
   END SUBROUTINE lim_dyn
#endif 

   !!======================================================================
END MODULE limdyn
